"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
/**
 * Returns a PValue as a flat object with string keys and PCollection values.
 *
 * The full set of PCollections reachable by this PValue will be returned,
 * with keys corresponding roughly to the path taken to get there.
 */
function flattenPValue(pValue, prefix = "") {
    const result = {};
    if (pValue === null || pValue === undefined) {
        // pass
    }
    else if (pValue instanceof Root) {
        // pass
    }
    else if (pValue instanceof PCollection) {
        if (prefix) {
            result[prefix] = pValue;
        }
        else {
            result.main = pValue;
        }
    }
    else {
        if (prefix) {
            prefix += ".";
        }
        if (pValue instanceof Array) {
            for (var i = 0; i < pValue.length; i++) {
                Object.assign(result, flattenPValue(pValue[i], prefix + i));
            }
        }
        else {
            for (const [key, subValue] of Object.entries(pValue)) {
                Object.assign(result, flattenPValue(subValue, prefix + key));
            }
        }
    }
    return result;
}
flattenPValue.__closure = () => ({ Root, PCollection, Array, Object });
/**
 * Wraps a PValue in a single object such that a transform can be applied to it.
 *
 * For example, Flatten takes a PCollection[] as input, but Array has no
 * apply(PTransformClass) method, so one writes
 *
 *    P([pcA, pcB, pcC]).apply(new Flatten())
 */
function P(pvalue) {
    return new PValueWrapper(pvalue);
}
P.__closure = () => ({ PValueWrapper });
Object.defineProperty(exports, "__esModule", { value: true });
exports.P = exports.flattenPValue = exports.PCollection = exports.Root = void 0;
const _tct_flatten_destructured_imports_1 = __importStar(require("./internal/pipeline"));
var Pipeline = _tct_flatten_destructured_imports_1.Pipeline;
const _tct_flatten_destructured_imports_2 = __importStar(require("./transforms/transform"));
var PTransform = _tct_flatten_destructured_imports_2.PTransform;
var AsyncPTransform = _tct_flatten_destructured_imports_2.AsyncPTransform;
var PTransformClass = _tct_flatten_destructured_imports_2.PTransformClass;
var AsyncPTransformClass = _tct_flatten_destructured_imports_2.AsyncPTransformClass;
var extractName = _tct_flatten_destructured_imports_2.extractName;
var withName = _tct_flatten_destructured_imports_2.withName;
const _tct_flatten_destructured_imports_3 = __importStar(require("./transforms/pardo"));
var parDo = _tct_flatten_destructured_imports_3.parDo;
var DoFn = _tct_flatten_destructured_imports_3.DoFn;
const _tct_flatten_destructured_imports_4 = __importStar(require("./proto/beam_runner_api"));
var runnerApi = _tct_flatten_destructured_imports_4;
/**
 * The base object on which one can start building a Beam DAG.
 * Generally followed by a source-like transform such as a read or impulse.
 */
class Root {
    constructor(pipeline) {
        this.pipeline = pipeline;
    }
    apply(transform) {
        if (!(transform instanceof PTransformClass)) {
            transform = new PTransformClassFromCallable(transform);
        }
        return this.pipeline.applyTransform(transform, this);
    }
    async asyncApply(transform) {
        if (!(transform instanceof AsyncPTransformClass)) {
            transform = new AsyncPTransformClassFromCallable(transform);
        }
        return await this.pipeline.asyncApplyTransform(transform, this);
    }
}
exports.Root = Root;
/**
 * A deferred, possibly distributed collection of elements.
 */
class PCollection {
    constructor(pipeline, id) {
        this.type = "pcollection";
        this.pipeline = pipeline;
        if (typeof id === "string") {
            this.id = id;
        }
        else {
            this.computeId = id;
        }
    }
    getId() {
        if (this.id === null || this.id === undefined) {
            this.id = this.computeId();
        }
        return this.id;
    }
    apply(transform) {
        if (!(transform instanceof PTransformClass)) {
            transform = new PTransformClassFromCallable(transform);
        }
        return this.pipeline.applyTransform(transform, this);
    }
    asyncApply(transform) {
        if (!(transform instanceof AsyncPTransformClass)) {
            transform = new AsyncPTransformClassFromCallable(transform);
        }
        return this.pipeline.asyncApplyTransform(transform, this);
    }
    map(fn, context = undefined) {
        var _tct_transform_1;
        return this.apply(withName("map(" + extractName(fn) + ")", parDo({
            process: (_tct_transform_1 = function* (element, context) {
                // While it's legal to call a function with extra arguments which will
                // be ignored, this can have surprising behavior (e.g. for map(console.log))
                yield context === null || context === undefined
                    ? fn(element)
                    : fn(element, context);
            }, _tct_transform_1.__closure = () => ({ fn }), _tct_transform_1),
        }, context)));
    }
    flatMap(fn, context = undefined) {
        var _tct_transform_2;
        return this.apply(withName("flatMap(" + extractName(fn) + ")", parDo({
            process: (_tct_transform_2 = function (element, context) {
                // While it's legal to call a function with extra arguments which will
                // be ignored, this can have surprising behavior (e.g. for map(console.log))
                return context === null || context === undefined
                    ? fn(element)
                    : fn(element, context);
            }, _tct_transform_2.__closure = () => ({ fn }), _tct_transform_2),
        }, context)));
    }
    root() {
        return new Root(this.pipeline);
    }
}
exports.PCollection = PCollection;
exports.flattenPValue = flattenPValue;
exports.P = P;
class PValueWrapper {
    constructor(pvalue) {
        this.pvalue = pvalue;
    }
    apply(transform, root = null) {
        if (!(transform instanceof PTransformClass)) {
            transform = new PTransformClassFromCallable(transform);
        }
        return this.pipeline(root).applyTransform(transform, this.pvalue);
    }
    async asyncApply(transform, root = null) {
        if (!(transform instanceof AsyncPTransformClass)) {
            transform = new AsyncPTransformClassFromCallable(transform);
        }
        return await this.pipeline(root).asyncApplyTransform(transform, this.pvalue);
    }
    pipeline(root = null) {
        if (root === null || root === undefined) {
            const flat = flattenPValue(this.pvalue);
            return Object.values(flat)[0].pipeline;
        }
        else {
            return root.pipeline;
        }
    }
}
class PTransformClassFromCallable extends PTransformClass {
    constructor(expander) {
        super(extractName(expander));
        this.expander = expander;
    }
    expandInternal(input, pipeline, transformProto) {
        return this.expander(input, pipeline, transformProto);
    }
}
class AsyncPTransformClassFromCallable extends AsyncPTransformClass {
    constructor(expander) {
        super(extractName(expander));
        this.expander = expander;
    }
    async asyncExpandInternal(input, pipeline, transformProto) {
        return this.expander(input, pipeline, transformProto);
    }
}
const _tct_flatten_destructured_imports_5 = __importStar(require("./serialization"));
var requireForSerialization = _tct_flatten_destructured_imports_5.requireForSerialization;
requireForSerialization("apache_beam.pvalue", exports);
//# sourceMappingURL=pvalue.js.map