"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
function globalRegistry() {
    return CODER_REGISTRY;
}
globalRegistry.__closure = () => ({ CODER_REGISTRY });
function writeByteCallback(val, buf, pos) {
    buf[pos] = val & 0xff;
}
/**
 * Write a single byte, as an unsigned integer, directly to the writer.
 */
function writeRawByte(b, writer) {
    writer._push?.(writeByteCallback, 1, b);
}
writeRawByte.__closure = () => ({ writeByteCallback });
function writeBytesCallback(val, buf, pos) {
    for (let i = 0; i < val.length; ++i) {
        buf[pos + i] = val[i];
    }
}
/**
 * Writes a sequence of bytes, as unsigned integers, directly to the writer,
 * without a prefixing with the length of the bytes that writer.bytes() does.
 */
function writeRawBytes(value, writer) {
    writer._push?.(writeBytesCallback, value.length, value);
}
writeRawBytes.__closure = () => ({ writeBytesCallback });
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeRawBytes = exports.writeRawByte = exports.Context = exports.globalRegistry = void 0;
const _tct_flatten_destructured_imports_1 = __importStar(require("protobufjs"));
var Writer = _tct_flatten_destructured_imports_1.Writer;
var Reader = _tct_flatten_destructured_imports_1.Reader;
const _tct_flatten_destructured_imports_2 = __importStar(require("../proto/beam_runner_api"));
var runnerApi = _tct_flatten_destructured_imports_2;
/**
 * A registry of coder classes.
 *
 * Coder classes are registered by their URN. Most coders have a URN of the type `"beam:coder:{coder_name}:v1"`,
 * and they can register a constructor that takes zero or more parameters as input.
 * Constructor input parameters for a coder are usually internal coders that represent sub-components
 * of the enclosing coder (e.g. a coder of key-value pairs (a.k.a. `KVCoder`) may receive the coder
 * for the key and the coder for the value as parameters).
 */
class CoderRegistry {
    constructor() {
        this.internal_registry = {};
    }
    getCoder(urn, payload = undefined, ...components) {
        const constructor = this.internal_registry[urn];
        if (constructor === undefined) {
            throw new Error("Could not find coder for URN " + urn);
        }
        if (payload && payload.length > 0) {
            return constructor(payload, ...components);
        }
        else {
            return constructor(...components);
        }
    }
    // TODO: Figure out how to branch on constructors (called with new) and
    // ordinary functions.
    register(urn, coderClass) {
        this.registerClass(urn, coderClass);
    }
    registerClass(urn, coderClass) {
        var _tct_transform_1;
        this.registerConstructor(urn, (_tct_transform_1 = (...args) => new coderClass(...args), _tct_transform_1.__closure = () => ({ coderClass }), _tct_transform_1));
    }
    registerConstructor(urn, constructor) {
        this.internal_registry[urn] = constructor;
    }
}
const CODER_REGISTRY = new CoderRegistry();
exports.globalRegistry = globalRegistry;
/**
 * The context for encoding a PCollection element.
 * For example, for strings of utf8 characters or bytes, `wholeStream` encoding means
 * that the string will be encoded as-is; while `needsDelimiter` encoding means that the
 * string will be encoded prefixed with its length.
 *
 * ```js
 * coder = new StrUtf8Coder()
 * w1 = new Writer()
 * coder.encode("my string", w, Context.wholeStream)
 * console.log(w1.finish())  // <= Prints the pure byte-encoding of the string
 * w2 = new Writer()
 * coder.encode("my string", w, Context.needsDelimiters)
 * console.log(w2.finish())  // <= Prints a length-prefix string of bytes
 * ```
 */
var Context;
(function (Context) {
    /**
     * Whole stream encoding/decoding means that the encoding/decoding function does not need to worry about
     * delimiting the start and end of the current element in the stream of bytes.
     */
    Context["wholeStream"] = "wholeStream";
    /**
     * Needs-delimiters encoding means that the encoding of data must be such that when decoding,
     * the coder is able to stop decoding data at the end of the current element.
     */
    Context["needsDelimiters"] = "needsDelimiters";
})(Context = exports.Context || (exports.Context = {}));
exports.writeRawByte = writeRawByte;
exports.writeRawBytes = writeRawBytes;
//# sourceMappingURL=coders.js.map