"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
function directRunner(options = {}) {
    return new DirectRunner(options);
}
directRunner.__closure = () => ({ DirectRunner });
/**
 * Rewrites the pipeline to be suitable for running as a single "bundle."
 *
 * Given a pipeline of the form
 *
 *    ...          ...
 *     |            |
 * mainPColl    sidePColl
 *     |            |
 *   parDo <-------/
 *     |
 *    ...
 *
 * Rewrites this to be in the form.
 *
 *    ...          ...
 *     |            |
 * mainPColl    sidePColl
 *     |            |
 *     |      collectSideOp
 *     |            |
 *     |        emptyPColl
 *     | <- - - - -/
 *     |
 *  bufferOp
 *     |
 * mainPcollCopy
 *     |
 *   parDo <----- sidePCollCopy
 *
 * Where collectSideOp collects its inputs for future retrieval and bufferOp
 * buffers all its inputs, invoking the downstream parDo only once finishBundle
 * is called. The dotted line from collectSideOp to bufferOp is not used to
 * pass any data, but acts as a control edge forcing bufferOp to be finished
 * only after all side inputs have been fully collected.
 *
 * A sidePCollCopy is also placed in the graph to allow for lookup of metadata
 * when executing the parDo operaton.
 */
function rewriteSideInputs(p, pipelineStateRef) {
    function uniqueName(container, base) {
        let name = base;
        let counter = 0;
        while (container[name]) {
            name = base + counter++;
        }
        return name;
    }
    p = runnerApi.Pipeline.clone(p);
    const pcolls = p.components.pcollections;
    const transforms = p.components.transforms;
    for (const [transformId, transform] of Object.entries(transforms)) {
        if (transform.spec?.urn === parDo.urn &&
            Object.keys(transform.inputs).length > 1) {
            const spec = runnerApi.ParDoPayload.fromBinary(transform.spec.payload);
            // Figure out which input is the main input.
            // TODO: (Typescript) Is there a clean way to do a set difference?
            // There is a proposal for Set.prototype.intersect: https://github.com/tc39/proposal-set-methods
            // But I don't think there is a nicer way at the moment.
            let mainPCollTag = undefined;
            for (const tag of Object.keys(transform.inputs)) {
                if (!spec.sideInputs[tag]) {
                    if (mainPCollTag)
                        throw new Error("Multiple non-side inputs.");
                    mainPCollTag = tag;
                }
            }
            if (!mainPCollTag)
                throw new Error("Missing non-side input.");
            // Add the extra logic for each of the side inputs.
            const bufferInputs = { mainPCollTag: transform.inputs[mainPCollTag] };
            for (const side of Object.keys(spec.sideInputs)) {
                const sidePCollId = transform.inputs[side];
                const sideCopyId = uniqueName(pcolls, sidePCollId + "-" + side);
                pcolls[sideCopyId] = pcolls[sidePCollId];
                transform.inputs[side] = sideCopyId;
                const controlPCollId = uniqueName(pcolls, sidePCollId + "-" + side + "-control");
                pcolls[controlPCollId] = pcolls[transform.inputs[mainPCollTag]];
                bufferInputs[side] = controlPCollId;
                const collectTransformId = uniqueName(transforms, transformId + "-" + side + "-collect");
                transforms[collectTransformId] = runnerApi.PTransform.create({
                    spec: {
                        urn: CollectSideOperator.urn,
                        payload: serializeFn({
                            transformId: transformId,
                            sideInputId: side,
                            accessPattern: spec.sideInputs[side].accessPattern.urn,
                            pipelineStateRef: pipelineStateRef,
                        }),
                    },
                    inputs: { input: sidePCollId },
                    outputs: { out: controlPCollId },
                });
            }
            // Add the buffering operator.
            const bufferId = uniqueName(transforms, transformId + "-buffer");
            const bufferOutputId = uniqueName(pcolls, bufferId + "-out");
            pcolls[bufferOutputId] = pcolls[mainPCollTag];
            transforms[bufferId] = runnerApi.PTransform.create({
                spec: { urn: BufferOperator.urn, payload: new Uint8Array() },
                inputs: bufferInputs,
                outputs: { out: bufferOutputId },
            });
            transform.inputs[mainPCollTag] = bufferOutputId;
        }
    }
    return p;
}
rewriteSideInputs.__closure = () => ({ runnerApi, Object, parDo, Error, CollectSideOperator, serializeFn, BufferOperator, Uint8Array });
/////
function onlyElement(arg) {
    if (arg.length > 1) {
        Error("Expecting exactly one element.");
    }
    return arg[0];
}
onlyElement.__closure = () => ({ Error });
Object.defineProperty(exports, "__esModule", { value: true });
exports.directRunner = void 0;
const _tct_flatten_destructured_imports_1 = __importStar(require("uuid"));
var uuid = _tct_flatten_destructured_imports_1;
const _tct_flatten_destructured_imports_2 = __importStar(require("protobufjs"));
var protobufjs = _tct_flatten_destructured_imports_2;
const long_1 = __importStar(require("long")), _tct_flatten_destructured_imports_3 = long_1;
var Long = _tct_flatten_destructured_imports_3.default;
const _tct_flatten_destructured_imports_4 = __importStar(require("../proto/beam_runner_api"));
var runnerApi = _tct_flatten_destructured_imports_4;
const _tct_flatten_destructured_imports_5 = __importStar(require("../proto/beam_fn_api"));
var fnApi = _tct_flatten_destructured_imports_5;
const _tct_flatten_destructured_imports_6 = __importStar(require("../proto/beam_runner_api"));
var PTransform = _tct_flatten_destructured_imports_6.PTransform;
var PCollection = _tct_flatten_destructured_imports_6.PCollection;
const _tct_flatten_destructured_imports_7 = __importStar(require("../proto/beam_fn_api"));
var ProcessBundleDescriptor = _tct_flatten_destructured_imports_7.ProcessBundleDescriptor;
const _tct_flatten_destructured_imports_8 = __importStar(require("../proto/beam_job_api"));
var JobState_Enum = _tct_flatten_destructured_imports_8.JobState_Enum;
const _tct_flatten_destructured_imports_9 = __importStar(require("../internal/pipeline"));
var Pipeline = _tct_flatten_destructured_imports_9.Pipeline;
const _tct_flatten_destructured_imports_10 = __importStar(require("../pvalue"));
var Root = _tct_flatten_destructured_imports_10.Root;
const _tct_flatten_destructured_imports_11 = __importStar(require("../transforms/internal"));
var impulse = _tct_flatten_destructured_imports_11.impulse;
var groupByKey = _tct_flatten_destructured_imports_11.groupByKey;
const _tct_flatten_destructured_imports_12 = __importStar(require("./runner"));
var Runner = _tct_flatten_destructured_imports_12.Runner;
var PipelineResult = _tct_flatten_destructured_imports_12.PipelineResult;
const _tct_flatten_destructured_imports_13 = __importStar(require("../worker/worker"));
var worker = _tct_flatten_destructured_imports_13;
const _tct_flatten_destructured_imports_14 = __importStar(require("../worker/operators"));
var operators = _tct_flatten_destructured_imports_14;
const _tct_flatten_destructured_imports_15 = __importStar(require("../worker/pardo_context"));
var createStateKey = _tct_flatten_destructured_imports_15.createStateKey;
const _tct_flatten_destructured_imports_16 = __importStar(require("../worker/metrics"));
var metrics = _tct_flatten_destructured_imports_16;
const _tct_flatten_destructured_imports_17 = __importStar(require("../worker/state"));
var state = _tct_flatten_destructured_imports_17;
const _tct_flatten_destructured_imports_18 = __importStar(require("../transforms/pardo"));
var parDo = _tct_flatten_destructured_imports_18.parDo;
const _tct_flatten_destructured_imports_19 = __importStar(require("../values"));
var Window = _tct_flatten_destructured_imports_19.Window;
var GlobalWindow = _tct_flatten_destructured_imports_19.GlobalWindow;
var Instant = _tct_flatten_destructured_imports_19.Instant;
var PaneInfo = _tct_flatten_destructured_imports_19.PaneInfo;
var WindowedValue = _tct_flatten_destructured_imports_19.WindowedValue;
const _tct_flatten_destructured_imports_20 = __importStar(require("../coders/standard_coders"));
var PaneInfoCoder = _tct_flatten_destructured_imports_20.PaneInfoCoder;
const _tct_flatten_destructured_imports_21 = __importStar(require("../coders/coders"));
var Coder = _tct_flatten_destructured_imports_21.Coder;
var CoderContext = _tct_flatten_destructured_imports_21.Context;
const _tct_flatten_destructured_imports_22 = __importStar(require("../internal/environments"));
var environments = _tct_flatten_destructured_imports_22;
const _tct_flatten_destructured_imports_23 = __importStar(require("../internal/serialize"));
var serializeFn = _tct_flatten_destructured_imports_23.serializeFn;
var deserializeFn = _tct_flatten_destructured_imports_23.deserializeFn;
const SUPPORTED_REQUIREMENTS = [];
exports.directRunner = directRunner;
class DirectRunner extends Runner {
    constructor(options = {}) {
        super();
        this.options = options;
    }
    unsupportedFeatures(pipeline, options = {}) {
        return [...this.unsupportedFeaturesIter(pipeline, options)];
    }
    *unsupportedFeaturesIter(proto, options = {}) {
        for (const requirement of proto.requirements) {
            if (!SUPPORTED_REQUIREMENTS.includes(requirement)) {
                yield requirement;
            }
        }
        for (const env of Object.values(proto.components.environments)) {
            if (env.urn &&
                env.urn !== environments.TYPESCRIPT_DEFAULT_ENVIRONMENT_URN) {
                yield env.urn;
            }
        }
        for (const windowing of Object.values(proto.components.windowingStrategies)) {
            if (![
                runnerApi.MergeStatus_Enum.UNSPECIFIED,
                runnerApi.MergeStatus_Enum.NON_MERGING,
                runnerApi.MergeStatus_Enum.ALREADY_MERGED,
            ].includes(windowing.mergeStatus)) {
                yield "MergeStatus=" + windowing.mergeStatus;
            }
            if (windowing.outputTime !== runnerApi.OutputTime_Enum.END_OF_WINDOW) {
                yield "OutputTime=" + windowing.outputTime;
            }
        }
    }
    async runPipeline(p) {
        const stateProvider = new InMemoryStateProvider();
        const stateCacheRef = uuid.v4();
        DirectRunner.inMemoryStatesRefs.set(stateCacheRef, stateProvider);
        try {
            const proto = rewriteSideInputs(p, stateCacheRef);
            const descriptor = {
                id: "",
                transforms: proto.components.transforms,
                pcollections: proto.components.pcollections,
                windowingStrategies: proto.components.windowingStrategies,
                coders: proto.components.coders,
                environments: proto.components.environments,
            };
            const processor = new worker.BundleProcessor(descriptor, null, new state.CachingStateProvider(stateProvider), [impulse.urn]);
            await processor.process("bundle_id");
            return new (class DirectPipelineResult extends PipelineResult {
                waitUntilFinish(duration) {
                    return Promise.resolve(JobState_Enum.DONE);
                }
                async rawMetrics() {
                    var _tct_transform_1;
                    const shortIdCache = new metrics.MetricsShortIdCache();
                    const monitoringData = processor.monitoringData(shortIdCache);
                    return Array.from(monitoringData.entries()).map((_tct_transform_1 = ([id, payload]) => shortIdCache.asMonitoringInfo(id, payload), _tct_transform_1.__closure = () => ({ shortIdCache }), _tct_transform_1));
                }
            })();
        }
        finally {
            DirectRunner.inMemoryStatesRefs.delete(stateCacheRef);
        }
    }
}
// All the operators for a given pipeline should share the same state.
// This global mapping allows operators to look up a shared state object for
// a given pipeline on deserialization.
DirectRunner.inMemoryStatesRefs = new Map();
// Only to be used in direct runner, as this will fire an element per worker, not per pipeline.
class DirectImpulseOperator {
    constructor(transformId, transform, context) {
        this.transformId = transformId;
        this.receiver = context.getReceiver(onlyElement(Object.values(transform.outputs)));
    }
    process(wvalue) {
        return operators.NonPromise;
    }
    async startBundle() {
        this.receiver.receive({
            value: new Uint8Array(),
            windows: [new GlobalWindow()],
            pane: PaneInfoCoder.ONE_AND_ONLY_FIRING,
            timestamp: Long.fromValue("-9223372036854775"), // TODO: (Cleanup) Pull constant out of proto, or at least as a constant elsewhere.
        });
    }
    async finishBundle() { }
}
operators.registerOperator(impulse.urn, DirectImpulseOperator);
// Only to be used in direct runner, as this will only group within a single bundle.
// TODO: (Extension) This could be used as a base for the PGBKOperation operator,
// and then this class could simply invoke that with an unbounded size and the
// concat-to-list CombineFn.
class DirectGbkOperator {
    constructor(transformId, transform, context) {
        this.transformId = transformId;
        this.receiver = context.getReceiver(onlyElement(Object.values(transform.outputs)));
        const inputPc = context.descriptor.pcollections[onlyElement(Object.values(transform.inputs))];
        this.keyCoder = context.pipelineContext.getCoder(context.descriptor.coders[inputPc.coderId].componentCoderIds[0]);
        const windowingStrategy = context.descriptor.windowingStrategies[inputPc.windowingStrategyId];
        if (windowingStrategy.mergeStatus !== runnerApi.MergeStatus_Enum.NON_MERGING) {
            throw new Error("Unsupported non-merging WindowFn: " + windowingStrategy);
        }
        if (windowingStrategy.outputTime !== runnerApi.OutputTime_Enum.END_OF_WINDOW) {
            throw new Error("Unsupported windowing output time: " + windowingStrategy);
        }
        this.windowCoder = context.pipelineContext.getCoder(windowingStrategy.windowCoderId);
    }
    process(wvalue) {
        for (const window of wvalue.windows) {
            const wkey = operators.encodeToBase64(window, this.windowCoder) +
                " " +
                operators.encodeToBase64(wvalue.value.key, this.keyCoder);
            if (!this.groups.has(wkey)) {
                this.groups.set(wkey, []);
            }
            this.groups.get(wkey).push(wvalue.value.value);
        }
        return operators.NonPromise;
    }
    async startBundle() {
        this.groups = new Map();
    }
    async finishBundle() {
        for (const [wkey, values] of this.groups) {
            const parts = wkey.split(" ");
            const encodedWindow = parts[0];
            const encodedKey = parts[1];
            const window = operators.decodeFromBase64(encodedWindow, this.windowCoder);
            const maybePromise = this.receiver.receive({
                value: {
                    key: operators.decodeFromBase64(encodedKey, this.keyCoder),
                    value: values,
                },
                windows: [window],
                timestamp: window.maxTimestamp(),
                pane: PaneInfoCoder.ONE_AND_ONLY_FIRING,
            });
            if (maybePromise !== operators.NonPromise) {
                await maybePromise;
            }
        }
        this.groups = null;
    }
}
operators.registerOperator(groupByKey.urn, DirectGbkOperator);
class CollectSideOperator {
    constructor(transformId, transform, context) {
        this.transformId = transformId;
        this.receiver = context.getReceiver(onlyElement(Object.values(transform.outputs)));
        const payload = deserializeFn(transform.spec.payload);
        this.parDoTransformId = payload.transformId;
        this.accessPattern = payload.accessPattern;
        this.sideInputId = payload.sideInputId;
        this.stateProvider = DirectRunner.inMemoryStatesRefs.get(payload.pipelineStateRef);
        const inputPc = context.descriptor.pcollections[onlyElement(Object.values(transform.inputs))];
        this.elementCoder = context.pipelineContext.getCoder(inputPc.coderId);
        const windowingStrategy = context.descriptor.windowingStrategies[inputPc.windowingStrategyId];
        this.windowCoder = context.pipelineContext.getCoder(windowingStrategy.windowCoderId);
    }
    async startBundle() { }
    process(wvalue) {
        for (const window of wvalue.windows) {
            const writer = new protobufjs.Writer();
            this.elementCoder.encode(wvalue.value, writer, CoderContext.needsDelimiters);
            const encodedElement = writer.finish();
            this.stateProvider.appendState(createStateKey(this.parDoTransformId, this.accessPattern, this.sideInputId, window, this.windowCoder), encodedElement);
        }
        return operators.NonPromise;
    }
    async finishBundle() { }
}
CollectSideOperator.urn = "beam:transforms:node_ts_direct:collect_side:v1";
operators.registerOperator(CollectSideOperator.urn, CollectSideOperator);
class BufferOperator {
    constructor(transformId, transform, context) {
        this.transformId = transformId;
        this.receiver = context.getReceiver(onlyElement(Object.values(transform.outputs)));
    }
    process(wvalue) {
        this.elements.push(wvalue);
        return operators.NonPromise;
    }
    async startBundle() {
        this.elements = [];
    }
    async finishBundle() {
        for (const element of this.elements) {
            const maybePromise = this.receiver.receive(element);
            if (maybePromise !== operators.NonPromise) {
                await maybePromise;
            }
        }
    }
}
BufferOperator.urn = "beam:transforms:node_ts_direct:buffer:v1";
operators.registerOperator(BufferOperator.urn, BufferOperator);
class InMemoryStateProvider {
    constructor() {
        this.chunks = new Map();
    }
    getState(stateKey, decode) {
        return {
            type: "value",
            value: decode(state.Uint8ArrayConcat(this.getStateEntry(stateKey))),
        };
    }
    appendState(stateKey, encodedData) {
        this.getStateEntry(stateKey).push(encodedData);
    }
    getStateEntry(stateKey) {
        const cacheKey = Buffer.from(fnApi.StateKey.toBinary(stateKey)).toString("base64");
        if (!this.chunks.has(cacheKey)) {
            this.chunks.set(cacheKey, []);
        }
        return this.chunks.get(cacheKey);
    }
}
//# sourceMappingURL=direct_runner.js.map