"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
/**
 * Creates a `Runner` object with the given set of options.
 *
 * The exact type of runner to be created can be specified by the special
 * `runner` option, e.g. `createRunner({runner: "direct"})` would create
 * a direct runner.  If no runner option is specified, the "default" runner
 * is used, which runs what pipelines it can on the direct runner, and
 * otherwise falls back to the universal runner (e.g. if cross-language
 * transforms, non-trivial windowing, etc. are used).
 */
function createRunner(options = {}) {
    let runnerConstructor;
    if (options.runner === undefined || options.runner === "default") {
        runnerConstructor = defaultRunner;
    }
    else if (options.runner === "direct") {
        runnerConstructor = require("./direct_runner").directRunner;
    }
    else if (options.runner === "universal") {
        runnerConstructor = require("./universal").universalRunner;
    }
    else if (options.runner === "flink") {
        runnerConstructor = require("./flink").flinkRunner;
    }
    else if (options.runner === "dataflow") {
        runnerConstructor = require("./dataflow").dataflowRunner;
    }
    else {
        throw new Error("Unknown runner: " + options.runner);
    }
    return runnerConstructor(options);
}
createRunner.__closure = () => ({ defaultRunner, require, Error });
function defaultRunner(defaultOptions) {
    return new (class extends Runner {
        async runPipeline(pipeline, options = {}) {
            const directRunner = require("./direct_runner").directRunner(defaultOptions);
            if (directRunner.unsupportedFeatures(pipeline, options).length === 0) {
                return directRunner.runPipeline(pipeline, options);
            }
            else {
                return loopbackRunner(defaultOptions).runPipeline(pipeline, options);
            }
        }
    })();
}
defaultRunner.__closure = () => ({ runPipeline, pipeline, options, require, loopbackRunner });
function loopbackRunner(defaultOptions = {}) {
    return require("./universal").universalRunner({
        environmentType: "LOOPBACK",
        ...defaultOptions,
    });
}
loopbackRunner.__closure = () => ({ require });
Object.defineProperty(exports, "__esModule", { value: true });
exports.loopbackRunner = exports.defaultRunner = exports.Runner = exports.createRunner = exports.PipelineResult = void 0;
const _tct_flatten_destructured_imports_1 = __importStar(require("../proto/beam_job_api"));
var JobState_Enum = _tct_flatten_destructured_imports_1.JobState_Enum;
const _tct_flatten_destructured_imports_2 = __importStar(require("../proto/beam_runner_api"));
var runnerApi = _tct_flatten_destructured_imports_2;
const _tct_flatten_destructured_imports_3 = __importStar(require("../proto/metrics"));
var MonitoringInfo = _tct_flatten_destructured_imports_3.MonitoringInfo;
const _tct_flatten_destructured_imports_4 = __importStar(require("../internal/pipeline"));
var Pipeline = _tct_flatten_destructured_imports_4.Pipeline;
const _tct_flatten_destructured_imports_5 = __importStar(require("../pvalue"));
var Root = _tct_flatten_destructured_imports_5.Root;
var PValue = _tct_flatten_destructured_imports_5.PValue;
const _tct_flatten_destructured_imports_6 = __importStar(require("../options/pipeline_options"));
var PipelineOptions = _tct_flatten_destructured_imports_6.PipelineOptions;
const _tct_flatten_destructured_imports_7 = __importStar(require("../worker/metrics"));
var metrics = _tct_flatten_destructured_imports_7;
class PipelineResult {
    waitUntilFinish(duration) {
        throw new Error("NotImplemented");
    }
    async rawMetrics() {
        throw new Error("NotImplemented");
    }
    // TODO: Support filtering, slicing.
    async counters() {
        return Object.fromEntries(metrics.aggregateMetrics(await this.rawMetrics(), "beam:metric:user:sum_int64:v1"));
    }
    async distributions() {
        return Object.fromEntries(metrics.aggregateMetrics(await this.rawMetrics(), "beam:metric:user:distribution_int64:v1"));
    }
}
exports.PipelineResult = PipelineResult;
exports.createRunner = createRunner;
/**
 * A Runner is the object that takes a pipeline definition and actually
 * executes, e.g. locally or on a distributed system, by invoking its
 * `run` or `runAsync` method.
 *
 * Runners are generally created using the `createRunner` method in this
 * same module.
 */
class Runner {
    /**
     * Runs the transform.
     *
     * Resolves to an instance of PipelineResult when the pipeline completes.
     * Use runAsync() to execute the pipeline in the background.
     *
     * @param pipeline
     * @returns A PipelineResult
     */
    async run(pipeline, options) {
        const pipelineResult = await this.runAsync(pipeline, options);
        const finalState = await pipelineResult.waitUntilFinish();
        if (finalState != JobState_Enum.DONE) {
            // TODO: Grab the last/most severe error message?
            throw new Error("Job finished in state " + JobState_Enum[finalState]);
        }
        return pipelineResult;
    }
    /**
     * runAsync() is the asynchronous version of run(), does not wait until
     * pipeline finishes. Use the returned PipelineResult to query job
     * status.
     */
    async runAsync(pipeline, options) {
        const p = new Pipeline();
        await pipeline(new Root(p));
        return this.runPipeline(p.getProto());
    }
}
exports.Runner = Runner;
exports.defaultRunner = defaultRunner;
exports.loopbackRunner = loopbackRunner;
//# sourceMappingURL=runner.js.map