"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
function serviceProviderFromJavaGradleTarget(gradleTarget, args = undefined) {
    return _tct_transform_1 = async () => {
        let jar;
        const serviceInfo = serviceOverrideFor(gradleTarget);
        if (serviceInfo) {
            if (serviceInfo.match(/^[a-zA-Z0-9.]+:[0-9]+$/)) {
                return new ExternalService(serviceInfo);
            }
            else {
                jar = serviceInfo;
            }
        }
        else {
            jar = await JavaJarService.cachedJar(await JavaJarService.gradleToJar(gradleTarget));
        }
        return new JavaJarService(jar, args, gradleTarget);
    }, _tct_transform_1.__closure = () => ({ serviceOverrideFor, gradleTarget, ExternalService, JavaJarService, args }), _tct_transform_1;
}
var _tct_transform_1;
serviceProviderFromJavaGradleTarget.__closure = () => ({ serviceOverrideFor, ExternalService, JavaJarService });
/**
 * Allows one to specify alternatives for auto-started services by specifying
 * an environment variable. For example,
 *
 * export BEAM_SERVICE_OVERRIDES='{
 *   "python:apache_beam.runners.portability.expansion_service_main":
 *       "localhost:port",
 *   "python:*": "/path/to/dev/venv/bin/python",
 *   "sdks:java:extensions:sql:expansion-service:shadowJar": "/path/to/jar"
 * }'
 *
 * would use the address localhost:port any time the Python service at
 * apache_beam.runners.portability.expansion_service_main was requested,
 * use the binary at /path/to/dev/venv/bin/python for starting up all other
 * services, and use /path/to/jar when the gradle target
 * sdks:java:extensions:sql:expansion-service:shadowJar was requested.
 *
 * This is primarily for testing.
 */
function serviceOverrideFor(name) {
    if (process.env.BEAM_SERVICE_OVERRIDES) {
        return JSON.parse(process.env.BEAM_SERVICE_OVERRIDES)[name];
    }
}
serviceOverrideFor.__closure = () => ({ process, JSON });
function getBeamProjectRoot() {
    try {
        const projectRoot = path.dirname(findGitRoot(__dirname));
        if (fs.existsSync(path.join(projectRoot, "sdks", "typescript", "src", "apache_beam"))) {
            return projectRoot;
        }
        else {
            return undefined;
        }
    }
    catch (Error) {
        return undefined;
    }
}
getBeamProjectRoot.__closure = () => ({ path, findGitRoot, __dirname, fs });
Object.defineProperty(exports, "__esModule", { value: true });
exports.PythonService = exports.JavaJarService = exports.serviceProviderFromJavaGradleTarget = exports.SubprocessService = exports.ExternalService = void 0;
const fs = require("fs");
const https = require("https");
const os = require("os");
const net = require("net");
const path = require("path");
const childProcess = require("child_process");
const findGitRoot = require("find-git-root");
// TODO: (Typescript) Why can't the var above be used as a namespace?
const _tct_flatten_destructured_imports_1 = __importStar(require("child_process"));
var ChildProcess = _tct_flatten_destructured_imports_1.ChildProcess;
const _tct_flatten_destructured_imports_2 = __importStar(require("../version"));
var beamVersion = _tct_flatten_destructured_imports_2.version;
class ExternalService {
    constructor(address) {
        this.address = address;
        this.address = address;
    }
    async start() {
        return this.address;
    }
    async stop() { }
}
exports.ExternalService = ExternalService;
class SubprocessServiceCache {
    constructor(parent = undefined) {
        this.services = new Map();
        this.parent = parent;
    }
    put(key, service) {
        this.services.set(key, service);
    }
    get(key) {
        if (this.services.has(key)) {
            return this.services.get(key);
        }
        else if (this.parent) {
            return this.parent?.get(key);
        }
        else {
            return undefined;
        }
    }
    async stopAll() {
        await Promise.all([...this.services.values()].map((service) => {
            service.cached = false;
            return service.stop();
        }));
    }
}
class SubprocessService {
    constructor(cmd, args, name = undefined) {
        this.cmd = cmd;
        this.args = args;
        this.name = name || cmd;
    }
    static async freePort() {
        var _tct_transform_2;
        return new Promise((_tct_transform_2 = (resolve) => {
            var _tct_transform_3;
            const srv = net.createServer();
            srv.listen(0, (_tct_transform_3 = () => {
                var _tct_transform_4;
                const port = srv.address().port;
                srv.close((_tct_transform_4 = (_) => resolve(port), _tct_transform_4.__closure = () => ({ resolve, port }), _tct_transform_4));
            }, _tct_transform_3.__closure = () => ({ srv, resolve }), _tct_transform_3));
        }, _tct_transform_2.__closure = () => ({ net }), _tct_transform_2));
    }
    static createCache() {
        SubprocessService.cache = new SubprocessServiceCache(SubprocessService.cache);
        return this.cache;
    }
    key() {
        return this.cmd + "\0" + this.args.join("\0");
    }
    async start() {
        var _tct_transform_5, _tct_transform_6;
        if (SubprocessService.cache) {
            const started = SubprocessService.cache.get(this.key());
            if (started) {
                this.cached = true;
                return started.address;
            }
        }
        const host = "localhost";
        const port = (await SubprocessService.freePort()).toString();
        console.debug(this.cmd, this.args.map((_tct_transform_5 = (arg) => arg.replace("{{PORT}}", port), _tct_transform_5.__closure = () => ({ port }), _tct_transform_5)));
        this.process = childProcess.spawn(this.cmd, this.args.map((_tct_transform_6 = (arg) => arg.replace("{{PORT}}", port), _tct_transform_6.__closure = () => ({ port }), _tct_transform_6)), {
            stdio: "inherit",
        });
        try {
            console.debug(`Waiting for ${this.name} to be available on port ${port}.`);
            await this.portReady(port, host, 10000);
            console.debug(`Service ${this.name} available.`);
        }
        catch (error) {
            this.process.kill();
            throw error;
        }
        this.address = host + ":" + port;
        if (SubprocessService.cache) {
            SubprocessService.cache.put(this.key(), this);
            this.cached = true;
        }
        else {
            this.cached = false;
        }
        return this.address;
    }
    async stop() {
        if (this.cached) {
            return;
        }
        console.info(`Tearing down ${this.name}.`);
        this.address = undefined;
        this.process.kill();
    }
    async portReady(port, host, timeoutMs, iterMs = 100) {
        var _tct_transform_7, _tct_transform_8;
        const start = Date.now();
        let connected = { value: undefined };
        connected.value = false;
        while (!connected.value && Date.now() - start < timeoutMs) {
            if (this.process.exitCode) {
                throw new Error("Aborted with error " + this.process.exitCode);
            }
            await new Promise((_tct_transform_7 = (r) => setTimeout(r, iterMs), _tct_transform_7.__closure = () => ({ setTimeout, iterMs }), _tct_transform_7));
            try {
                await new Promise((_tct_transform_8 = (resolve, reject) => {
                    var _tct_transform_9, _tct_transform_10;
                    const socket = net.createConnection(port, host, (_tct_transform_9 = () => {
                        connected.value = true;
                        socket.end();
                        resolve();
                    }, _tct_transform_9.__closure = () => ({ connected, socket, resolve }), _tct_transform_9));
                    socket.on("error", (_tct_transform_10 = (err) => {
                        reject(err);
                    }, _tct_transform_10.__closure = () => ({ reject }), _tct_transform_10));
                }, _tct_transform_8.__closure = () => ({ net, port, host, connected }), _tct_transform_8));
            }
            catch (err) {
                // go around again
            }
        }
        if (!connected.value) {
            throw new Error("Timed out waiting for service after " + timeoutMs + "ms.");
        }
    }
}
exports.SubprocessService = SubprocessService;
SubprocessService.cache = undefined;
exports.serviceProviderFromJavaGradleTarget = serviceProviderFromJavaGradleTarget;
const BEAM_CACHE = path.join(os.homedir(), ".apache_beam", "cache");
class JavaJarService extends SubprocessService {
    constructor(jar, args = undefined, name = undefined) {
        if (!args) {
            // TODO: (Extension) Should filesToStage be set at some higher level?
            args = ["{{PORT}}", "--filesToStage=" + jar];
        }
        super("java", ["-jar", jar].concat(args), name);
    }
    static async cachedJar(urlOrPath, cacheDir = JavaJarService.JAR_CACHE) {
        var _tct_transform_11;
        if (urlOrPath.match(/^https?:\/\//)) {
            fs.mkdirSync(cacheDir, { recursive: true });
            const dest = path.join(JavaJarService.JAR_CACHE, path.basename(urlOrPath));
            if (fs.existsSync(dest)) {
                return dest;
            }
            // TODO: (Cleanup) Use true temporary file.
            const tmp = dest + ".tmp" + Math.random();
            return new Promise((_tct_transform_11 = (resolve, reject) => {
                var _tct_transform_12;
                const fout = fs.createWriteStream(tmp);
                console.warn("Downloading", urlOrPath);
                const request = https.get(urlOrPath, (_tct_transform_12 = function (response) {
                    var _tct_transform_13;
                    if (response.statusCode !== 200) {
                        reject(`Error code ${response.statusCode} when downloading ${urlOrPath}`);
                    }
                    response.pipe(fout);
                    fout.on("finish", (_tct_transform_13 = function () {
                        var _tct_transform_14;
                        fout.close((_tct_transform_14 = () => {
                            fs.renameSync(tmp, dest);
                            resolve(dest);
                        }, _tct_transform_14.__closure = () => ({ fs, tmp, dest, resolve }), _tct_transform_14));
                    }, _tct_transform_13.__closure = () => ({ fout, fs, tmp, dest, resolve }), _tct_transform_13));
                }, _tct_transform_12.__closure = () => ({ reject, urlOrPath, fout, fs, tmp, dest, resolve }), _tct_transform_12));
            }, _tct_transform_11.__closure = () => ({ fs, tmp, console, urlOrPath, https, dest }), _tct_transform_11));
        }
        else {
            return urlOrPath;
        }
    }
    static async gradleToJar(gradleTarget, appendix = undefined, version = beamVersion) {
        if (version.startsWith("0.")) {
            // node-ts 0.x corresponds to Beam 2.x.
            version = "2" + version.substring(1);
        }
        const gradlePackage = gradleTarget.match(/^:?(.*):[^:]+:?$/)[1];
        const artifactId = "beam-" + gradlePackage.replaceAll(":", "-");
        const projectRoot = getBeamProjectRoot();
        const localPath = !projectRoot
            ? undefined
            : path.join(projectRoot, gradlePackage.replaceAll(":", path.sep), "build", "libs", JavaJarService.jarName(artifactId, version.replace("-SNAPSHOT", ""), "SNAPSHOT", appendix));
        if (version.includes("SNAPSHOT") && !projectRoot) {
            version = "latest";
        }
        if (localPath && fs.existsSync(localPath)) {
            console.info("Using pre-built snapshot at", localPath);
            return localPath;
        }
        else if (version.includes("SNAPSHOT")) {
            throw new Error(`${localPath} not found. Please build the server with
      cd ${projectRoot}; ./gradlew ${gradleTarget})`);
        }
        else {
            return JavaJarService.mavenJarUrl(artifactId, version, undefined, appendix);
        }
    }
    static async mavenJarUrl(artifactId, version, classifier = undefined, appendix = undefined, repo = JavaJarService.APACHE_REPOSITORY, groupId = JavaJarService.BEAM_GROUP_ID) {
        var _tct_transform_15;
        if (version == "latest") {
            const medatadataUrl = [
                repo,
                groupId.replaceAll(".", "/"),
                artifactId,
                "maven-metadata.xml",
            ].join("/");
            const metadata = await new Promise((_tct_transform_15 = (resolve, reject) => {
                var _tct_transform_16;
                let data = { value: undefined };
                data.value = "";
                https.get(medatadataUrl, (_tct_transform_16 = (res) => {
                    var _tct_transform_17, _tct_transform_18, _tct_transform_19;
                    res.on("data", (_tct_transform_17 = (chunk) => {
                        data.value = data.value + chunk;
                    }, _tct_transform_17.__closure = () => ({ data }), _tct_transform_17));
                    res.on("end", (_tct_transform_18 = () => {
                        resolve(data.value);
                    }, _tct_transform_18.__closure = () => ({ resolve, data }), _tct_transform_18));
                    res.on("error", (_tct_transform_19 = (e) => {
                        reject(e);
                    }, _tct_transform_19.__closure = () => ({ reject }), _tct_transform_19));
                }, _tct_transform_16.__closure = () => ({ data, resolve, reject }), _tct_transform_16));
            }, _tct_transform_15.__closure = () => ({ https, medatadataUrl }), _tct_transform_15));
            version = metadata.match(/<latest>(.*)<\/latest>/)[1];
        }
        return [
            repo,
            groupId.replaceAll(".", "/"),
            artifactId,
            version,
            JavaJarService.jarName(artifactId, version, classifier, appendix),
        ].join("/");
    }
    static jarName(artifactId, version, classifier, appendix) {
        return ([artifactId, appendix, version, classifier]
            .filter((s) => s !== null && s !== undefined)
            .join("-") + ".jar");
    }
}
exports.JavaJarService = JavaJarService;
JavaJarService.APACHE_REPOSITORY = "https://repo.maven.apache.org/maven2";
JavaJarService.BEAM_GROUP_ID = "org.apache.beam";
JavaJarService.JAR_CACHE = path.join(BEAM_CACHE, "jars");
class PythonService extends SubprocessService {
    constructor(pythonExecutablePath, module, args = []) {
        super(pythonExecutablePath, ["-u", "-m", module].concat(args), module);
    }
    static whichPython() {
        for (const bin of ["python3", "python"]) {
            try {
                const result = childProcess.spawnSync(bin, ["--version"]);
                if (result.status === 0) {
                    return bin;
                }
            }
            catch (err) {
                // Try the next one.
            }
        }
        throw new Error("Can't find a Python executable.");
    }
    static beamPython() {
        const bootstrapScript = path.join(__dirname, "..", "..", "..", "resources", "bootstrap_beam_venv.py");
        console.debug("Invoking Python bootstrap script.");
        const result = childProcess.spawnSync(PythonService.whichPython(), [bootstrapScript], { encoding: "latin1" });
        if (result.status === 0) {
            console.debug(result.stdout);
            const lines = result.stdout.trim().split("\n");
            return lines[lines.length - 1];
        }
        else {
            throw new Error(result.output);
        }
    }
    static forModule(module, args = []) {
        let pythonExecutable;
        let serviceInfo = serviceOverrideFor("python:" + module);
        if (!serviceInfo) {
            serviceInfo = serviceOverrideFor("python:*");
        }
        if (serviceInfo) {
            if (serviceInfo.match(/^[a-zA-Z0-9.]+:[0-9]+$/)) {
                return new ExternalService(serviceInfo);
            }
            else {
                pythonExecutable = serviceInfo;
            }
        }
        else {
            pythonExecutable = PythonService.beamPython();
        }
        return new PythonService(pythonExecutable, module, args);
    }
}
exports.PythonService = PythonService;
PythonService.VENV_CACHE = path.join(BEAM_CACHE, "venvs");
//# sourceMappingURL=service.js.map