"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
function createMetric(spec) {
    const constructor = metricTypes.get(spec.metricType);
    if (constructor === undefined) {
        throw new Error("Unknown metric type: " + spec.metricType);
    }
    return constructor(spec);
}
var _tct_transform_1, _tct_transform_2, _tct_transform_3;
createMetric.__closure = () => ({ metricTypes, Error });
function aggregateMetrics(infos, urn) {
    const cells = new Map();
    for (const info of infos) {
        if (info.urn == urn) {
            const name = info.labels.NAME;
            const cell = createMetric({ metricType: info.urn, name });
            cell.loadFromPayload(info.payload);
            if (cells.has(name)) {
                cells.get(name).merge(cell);
            }
            else {
                cells.set(name, cell);
            }
        }
    }
    return new Map(Array.from(cells.entries()).map(([name, cell]) => [name, cell.extract()]));
}
aggregateMetrics.__closure = () => ({ Map, createMetric, Array });
function encode(value, coder) {
    const buffer = new protobufjs.Writer();
    coder.encode(value, buffer, Context.needsDelimiters);
    return buffer.finish();
}
encode.__closure = () => ({ protobufjs, Context });
function decode(encoded, coder) {
    return coder.decode(new protobufjs.Reader(encoded), Context.needsDelimiters);
}
decode.__closure = () => ({ protobufjs, Context });
Object.defineProperty(exports, "__esModule", { value: true });
exports.aggregateMetrics = exports.MetricsShortIdCache = exports.MetricsContainer = void 0;
const _tct_flatten_destructured_imports_1 = __importStar(require("protobufjs"));
var protobufjs = _tct_flatten_destructured_imports_1;
const _tct_flatten_destructured_imports_2 = __importStar(require("../proto/metrics"));
var MonitoringInfo = _tct_flatten_destructured_imports_2.MonitoringInfo;
const _tct_flatten_destructured_imports_3 = __importStar(require("../coders/coders"));
var Coder = _tct_flatten_destructured_imports_3.Coder;
var Context = _tct_flatten_destructured_imports_3.Context;
const _tct_flatten_destructured_imports_4 = __importStar(require("../coders/required_coders"));
var IterableCoder = _tct_flatten_destructured_imports_4.IterableCoder;
const _tct_flatten_destructured_imports_5 = __importStar(require("../coders/standard_coders"));
var VarIntCoder = _tct_flatten_destructured_imports_5.VarIntCoder;
const metricTypes = new Map();
class Counter {
    constructor(spec) {
        this.value = 0;
    }
    update(value) {
        this.value += value;
    }
    reset() {
        this.value = 0;
    }
    toEmptyMonitoringInfo() {
        return MonitoringInfo.create({
            urn: "beam:metric:user:sum_int64:v1",
            type: "beam:metrics:sum_int64:v1",
        });
    }
    payload() {
        return encode(this.value, new VarIntCoder());
    }
    loadFromPayload(payload) {
        this.value = decode(payload, new VarIntCoder());
    }
    merge(other) {
        this.value += other.value;
    }
    extract() {
        return this.value;
    }
}
metricTypes.set("beam:metric:user:sum_int64:v1", (_tct_transform_1 = (spec) => new Counter(spec), _tct_transform_1.__closure = () => ({ Counter }), _tct_transform_1));
class Distribution {
    constructor(spec) {
        this.count = 0;
        this.sum = 0;
        this.min = 0;
        this.max = 0;
        this.coder = new IterableCoder(new VarIntCoder());
    }
    update(value) {
        if (this.count == 0) {
            this.count = 1;
            this.sum = this.min = this.max = value;
        }
        else {
            this.count += 1;
            this.sum += value;
            this.min = Math.min(value, this.min);
            this.max = Math.max(value, this.max);
        }
    }
    reset() {
        this.count = 0;
    }
    toEmptyMonitoringInfo() {
        return MonitoringInfo.create({
            urn: "beam:metric:user:distribution_int64:v1",
            type: "beam:metrics:distribution_int64:v1",
        });
    }
    payload() {
        return encode([this.count, this.sum, this.min, this.max], this.coder);
    }
    loadFromPayload(payload) {
        [this.count, this.sum, this.min, this.max] = decode(payload, this.coder);
    }
    merge(other) {
        if (other.count == 0) {
            return;
        }
        else if (this.count == 0) {
            this.count = other.count;
            this.sum = other.sum;
            this.min = other.min;
            this.max = other.max;
        }
        else {
            this.count += other.count;
            this.sum += other.sum;
            this.min = Math.min(this.min, other.min);
            this.max = Math.max(this.max, other.max);
        }
    }
    extract() {
        if (this.count == 0) {
            return {
                count: 0,
                sum: 0,
                min: NaN,
                max: NaN,
            };
        }
        else {
            return {
                count: this.count,
                sum: this.sum,
                min: this.min,
                max: this.max,
            };
        }
    }
}
metricTypes.set("beam:metric:user:distribution_int64:v1", (_tct_transform_2 = (spec) => new Distribution(spec), _tct_transform_2.__closure = () => ({ Distribution }), _tct_transform_2));
class ElementCount extends Counter {
    toEmptyMonitoringInfo() {
        return MonitoringInfo.create({
            urn: "beam:metric:element_count:v1",
            type: "beam:metrics:sum_int64:v1",
        });
    }
}
metricTypes.set("beam:metric:user:element_count:v1", (_tct_transform_3 = (spec) => new ElementCount(spec), _tct_transform_3.__closure = () => ({ ElementCount }), _tct_transform_3));
/**
 * A ScopedMetricCell is a MetricCell together with its identifier(s).
 */
class ScopedMetricCell {
    constructor(key, value, name, transformId, pcollectionId) {
        this.key = key;
        this.value = value;
        this.name = name;
        this.transformId = transformId;
        this.pcollectionId = pcollectionId;
        this.is_set = false;
    }
    toEmptyMonitoringInfo() {
        const info = this.value.toEmptyMonitoringInfo();
        if (this.name) {
            info.labels["NAME"] = this.name;
            info.labels["NAMESPACE"] = "";
        }
        if (this.transformId) {
            info.labels["TRANSFORM"] = this.transformId;
        }
        if (this.pcollectionId) {
            info.labels["PCOLLECTION"] = this.pcollectionId;
        }
        return info;
    }
    update(value) {
        this.is_set = true;
        this.value.update(value);
    }
    reset() {
        this.is_set = false;
    }
}
/**
 * A MetricsContainer holds a set of metrics for a particular bundle.
 *
 * It can be re-used, but must be reset() between bundles.
 */
class MetricsContainer {
    constructor() {
        this.metrics = new Map();
    }
    getMetric(transformId, pcollectionId, spec) {
        const key = spec.metricType +
            (transformId ? transformId.length + transformId : "_") +
            (pcollectionId ? pcollectionId.length + pcollectionId : "_") +
            spec.name;
        var cell = this.metrics.get(key);
        if (cell === undefined) {
            cell = new ScopedMetricCell(key, createMetric(spec), spec.name, transformId, pcollectionId);
            this.metrics.set(key, cell);
        }
        return cell;
    }
    elementCountMetric(pcollectionId) {
        return this.getMetric(undefined, pcollectionId, {
            name: undefined,
            metricType: "beam:metric:user:element_count:v1",
        });
    }
    monitoringData(shortIdCache) {
        var _tct_transform_4;
        return new Map(Array.from(this.metrics.values())
            .filter((metric) => metric.is_set)
            .map((_tct_transform_4 = (metric) => [
            shortIdCache.getShortId(metric),
            metric.value.payload(),
        ], _tct_transform_4.__closure = () => ({ shortIdCache }), _tct_transform_4)));
    }
    reset() {
        for (const [_, cell] of this.metrics) {
            cell.reset();
        }
    }
}
exports.MetricsContainer = MetricsContainer;
class MetricsShortIdCache {
    constructor() {
        this.idToInfo = new Map();
        this.keyToId = new Map();
    }
    getShortId(metric) {
        var shortId = this.keyToId.get(metric.key);
        if (shortId === undefined) {
            shortId = "m" + this.keyToId.size;
            this.idToInfo.set(shortId, metric.toEmptyMonitoringInfo());
            this.keyToId.set(metric.key, shortId);
        }
        return shortId;
    }
    asMonitoringInfo(id, payload = undefined) {
        const result = this.idToInfo.get(id);
        if (payload !== undefined) {
            return { ...result, payload };
        }
        else {
            return result;
        }
    }
}
exports.MetricsShortIdCache = MetricsShortIdCache;
exports.aggregateMetrics = aggregateMetrics;
//# sourceMappingURL=metrics.js.map