"use strict";
/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
// TODO: (API) Do we need an AsyncDoFn (and async[Flat]Map) to be able to call
// async functions in the body of the fns. Or can they always be Async?
// The latter seems to have perf issues.
// (For PTransformClasss, it's a major usability issue, but maybe we can always
// await when calling user code.  OTOH, I don't know what the performance
// impact would be for creating promises for every element of every operation
// which is typically a very performance critical spot to optimize.)
// TODO: (Typescript) Can the context arg be optional iff ContextT is undefined?
function parDo(doFn, context = undefined) {
    function expandInternal(input, pipeline, transformProto) {
        // Extract and populate side inputs from the context.
        const sideInputs = {};
        var contextCopy;
        if (typeof context === "object") {
            contextCopy = Object.create(context);
            const components = pipeline.context.components;
            for (const [name, value] of Object.entries(context)) {
                if (value instanceof SideInputParam) {
                    const inputName = "side." + name;
                    transformProto.inputs[inputName] = value.pcoll.getId();
                    contextCopy[name] = copySideInputWithId(value, inputName);
                    const mainWindowingStrategyId = components.pcollections[input.getId()].windowingStrategyId;
                    const sideWindowingStrategyId = components.pcollections[transformProto.inputs[inputName]]
                        .windowingStrategyId;
                    const sideWindowingStrategy = components.windowingStrategies[sideWindowingStrategyId];
                    const isGlobalSide = sideWindowingStrategy.windowFn.urn ==
                        "beam:window_fn:global_windows:v1";
                    sideInputs[inputName] = {
                        accessPattern: {
                            urn: value.accessor.accessPattern,
                            payload: new Uint8Array(),
                        },
                        // TODO: (Cleanup) The viewFn is stored in the side input object.
                        // Unclear what benefit there is to putting it here.
                        viewFn: { urn: "unused", payload: new Uint8Array() },
                        // TODO: (Extension) Possibly place this in the accessor.
                        windowMappingFn: {
                            urn: isGlobalSide
                                ? urns.GLOBAL_WINDOW_MAPPING_FN_URN
                                : mainWindowingStrategyId === sideWindowingStrategyId
                                    ? urns.IDENTITY_WINDOW_MAPPING_FN_URN
                                    : urns.ASSIGN_MAX_TIMESTAMP_WINDOW_MAPPING_FN_URN,
                            value: new Uint8Array(),
                        },
                    };
                }
                else {
                    contextCopy[name] = value;
                }
            }
        }
        else {
            contextCopy = context;
        }
        // Now finally construct the proto.
        transformProto.spec = runnerApi.FunctionSpec.create({
            urn: parDo.urn,
            payload: runnerApi.ParDoPayload.toBinary(runnerApi.ParDoPayload.create({
                doFn: runnerApi.FunctionSpec.create({
                    urn: urns.SERIALIZED_JS_DOFN_INFO,
                    payload: serializeFn({
                        doFn: doFn,
                        context: contextCopy,
                    }),
                }),
                sideInputs: sideInputs,
            })),
        });
        // For the ParDo output coder, we use a GeneralObjectCoder, which is a Javascript-specific
        // coder to encode the various types that exist in JS.
        // TODO: (Types) Should there be a way to specify, or better yet infer, the coder to use?
        return pipeline.createPCollectionInternal(new GeneralObjectCoder());
    }
    expandInternal.__closure = () => ({ context, Object, SideInputParam, copySideInputWithId, Uint8Array, urns, runnerApi, parDo, serializeFn, doFn, GeneralObjectCoder });
    return withName(`parDo(${extractName(doFn)})`, expandInternal);
}
var _tct_transform_1, _tct_transform_2;
parDo.__closure = () => ({ Object, SideInputParam, copySideInputWithId, Uint8Array, urns, runnerApi, serializeFn, GeneralObjectCoder, withName, extractName });
/**
 * Splits a single PCollection of objects, with keys k, into an object of
 * PCollections, with the same keys k, where each PCollection consists of the
 * values associated with that key. That is,
 *
 * PCollection<{a: T, b: U, ...}> maps to {a: PCollection<T>, b: PCollection<U>, ...}
 */
// TODO: (API) Consider as top-level method.
// TODO: Naming.
function split(tags, options = {}) {
    function expandInternal(input, pipeline, transformProto) {
        if (options.exclusive === undefined) {
            options.exclusive = true;
        }
        if (options.unknownTagBehavior === undefined) {
            options.unknownTagBehavior = "error";
        }
        if (options.unknownTagBehavior === "rename" &&
            !tags.includes(options.unknownTagName)) {
            tags.push(options.unknownTagName);
        }
        if (options.knownTags === undefined) {
            options.knownTags = tags;
        }
        transformProto.spec = runnerApi.FunctionSpec.create({
            urn: parDo.urn,
            payload: runnerApi.ParDoPayload.toBinary(runnerApi.ParDoPayload.create({
                doFn: runnerApi.FunctionSpec.create({
                    urn: urns.SPLITTING_JS_DOFN_URN,
                    payload: serializeFn(options),
                }),
            })),
        });
        return Object.fromEntries(tags.map((_tct_transform_1 = (tag) => [
            tag,
            pipeline.createPCollectionInternal(pipeline.context.getPCollectionCoderId(input)),
        ], _tct_transform_1.__closure = () => ({ pipeline, input }), _tct_transform_1)));
    }
    expandInternal.__closure = () => ({ options, tags, runnerApi, parDo, urns, serializeFn, Object });
    return withName(`Split(${tags})`, expandInternal);
}
split.__closure = () => ({ runnerApi, parDo, urns, serializeFn, Object, withName });
function windowParam() {
    return new ParDoParam("window");
}
windowParam.__closure = () => ({ ParDoParam });
function timestampParam() {
    return new ParDoParam("timestamp");
}
timestampParam.__closure = () => ({ ParDoParam });
// TODO: Naming. Should this be PaneParam?
function paneInfoParam() {
    return new ParDoParam("paneinfo");
}
paneInfoParam.__closure = () => ({ ParDoParam });
function copySideInputWithId(sideInput, id) {
    const copy = Object.create(sideInput);
    copy.sideInputId = id;
    delete copy.pcoll;
    return copy;
}
copySideInputWithId.__closure = () => ({ Object });
function iterableSideInput(pcoll) {
    return new SideInputParam(pcoll, {
        accessPattern: "beam:side_input:iterable:v1",
        toValue: (iter) => iter,
    });
}
iterableSideInput.__closure = () => ({ SideInputParam });
function singletonSideInput(pcoll, defaultValue = undefined) {
    return new SideInputParam(pcoll, {
        accessPattern: "beam:side_input:iterable:v1",
        toValue: (_tct_transform_2 = (iter) => {
            const asArray = Array.from(iter);
            if (asArray.length === 0 &&
                defaultValue !== null &&
                defaultValue !== undefined) {
                return defaultValue;
            }
            else if (asArray.length === 1) {
                return asArray[0];
            }
            else {
                throw new Error("Expected a single element, got " + asArray.length);
            }
        }, _tct_transform_2.__closure = () => ({ Array, defaultValue, Error }), _tct_transform_2),
    });
}
singletonSideInput.__closure = () => ({ SideInputParam, Array, Error });
Object.defineProperty(exports, "__esModule", { value: true });
exports.singletonSideInput = exports.iterableSideInput = exports.SideInputParam = exports.paneInfoParam = exports.timestampParam = exports.windowParam = exports.ParDoParam = exports.split = exports.parDo = void 0;
const _tct_flatten_destructured_imports_1 = __importStar(require("../proto/beam_runner_api"));
var runnerApi = _tct_flatten_destructured_imports_1;
const _tct_flatten_destructured_imports_2 = __importStar(require("../internal/urns"));
var urns = _tct_flatten_destructured_imports_2;
const _tct_flatten_destructured_imports_3 = __importStar(require("../coders/js_coders"));
var GeneralObjectCoder = _tct_flatten_destructured_imports_3.GeneralObjectCoder;
const _tct_flatten_destructured_imports_4 = __importStar(require("../pvalue"));
var PCollection = _tct_flatten_destructured_imports_4.PCollection;
const _tct_flatten_destructured_imports_5 = __importStar(require("../internal/pipeline"));
var Pipeline = _tct_flatten_destructured_imports_5.Pipeline;
const _tct_flatten_destructured_imports_6 = __importStar(require("../internal/serialize"));
var serializeFn = _tct_flatten_destructured_imports_6.serializeFn;
const _tct_flatten_destructured_imports_7 = __importStar(require("./transform"));
var PTransform = _tct_flatten_destructured_imports_7.PTransform;
var PTransformClass = _tct_flatten_destructured_imports_7.PTransformClass;
var withName = _tct_flatten_destructured_imports_7.withName;
var extractName = _tct_flatten_destructured_imports_7.extractName;
const _tct_flatten_destructured_imports_8 = __importStar(require("../values"));
var PaneInfo = _tct_flatten_destructured_imports_8.PaneInfo;
var Instant = _tct_flatten_destructured_imports_8.Instant;
var Window = _tct_flatten_destructured_imports_8.Window;
var WindowedValue = _tct_flatten_destructured_imports_8.WindowedValue;
exports.parDo = parDo;
// TODO: (Cleanup) use runnerApi.StandardPTransformClasss_Primitives.PAR_DO.urn.
parDo.urn = "beam:transform:pardo:v1";
exports.split = split;
/*
 * This is the root class of special parameters that can be provided in the
 * context of a map or DoFn.process method.  At runtime, one can invoke the
 * special `lookup` method to retrieve the relevant value associated with the
 * currently-being-processed element.
 */
class ParDoParam {
    constructor(parDoParamName) {
        this.parDoParamName = parDoParamName;
    }
    // TODO: Nameing "get" seems to be special.
    lookup() {
        if (this.provider === undefined) {
            throw new Error("Cannot be called outside of a DoFn's process method.");
        }
        return this.provider.provide(this);
    }
}
exports.ParDoParam = ParDoParam;
exports.windowParam = windowParam;
exports.timestampParam = timestampParam;
exports.paneInfoParam = paneInfoParam;
// TODO: (Extension) Support side inputs that are composites of multiple more
// primitive side inputs.
class SideInputParam extends ParDoParam {
    constructor(pcoll, accessor) {
        super("sideInput");
        this.pcoll = pcoll;
        this.accessor = accessor;
    }
}
exports.SideInputParam = SideInputParam;
exports.iterableSideInput = iterableSideInput;
exports.singletonSideInput = singletonSideInput;
// TODO: (Extension) Map side inputs.
// TODO: (Extension) Add providers for state, timers,
// restriction trackers, counters, etc.
const _tct_flatten_destructured_imports_9 = __importStar(require("../serialization"));
var requireForSerialization = _tct_flatten_destructured_imports_9.requireForSerialization;
requireForSerialization("apache_beam.transforms.pardo", exports);
//# sourceMappingURL=pardo.js.map